require "Window"

local KiruBox = {} 

local Type_Basic = "Basic"
local Type_Main = "Standard"

local versionNumber = "0.96"
 
function KiruBox:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 

	self.Preferences = {}  --preferences
	self.Preferences.FromRestore = {} --to hold restored settings
	self.Windows = {} --hold window spawns

    return o
end

function KiruBox:Init()
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
		-- "UnitOrPackageName",
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end
 

function KiruBox:OnLoad()
	self.xmlDoc = XmlDoc.CreateFromFile("KiruBox.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	
	Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
	Apollo.RegisterEventHandler("NewKiruBox", "OnKiruBoxOn", self)	
end

function KiruBox:OnInterfaceMenuListHasLoaded()
	Event_FireGenericEvent("InterfaceMenuList_NewAddOn", "KiruBox", {"NewKiruBox", "", ""})
end

function KiruBox:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		--process preferences for type of window
		local initialType = Type_Main
		local loadFirst = true
		
		if self.FromRestore ~= nil then 
			self.Preferences.WindowType = self.FromRestore.WindowType
			if self.FromRestore.Windows ~= nil and self.FromRestore.Windows[1] ~= nil then initialType = self.FromRestore.Windows[1].Type end
			if self.FromRestore.LoadFirst ~= nil then loadFirst = self.FromRestore.LoadFirst end
			self.Preferences.WindowType = self.FromRestore.WindowType or Type_Main
		end	
		if self.Preferences.WindowType == nil then self.Preferences.WindowType = Type_Main end
		if initialType == nil then initialType = self.Preferences.WindowType end
										
		--populate preferences and defaults
		if self.FromRestore ~= nil then
			self.Continue = self.FromRestore.ContinueText or "(continued)"
			self.Prefix  = self.FromRestore.PrefixText or ""
			self.Preferences.WindowPos = self.FromRestore.Windows
		else
			self.Continue = "(continued)"
			self.Prefix = ""
		end
		
		self.ContinueLen = #self.Continue
		self.PrefixLen = #self.Prefix
		
		self.RPPrefix = "{*}"
		
		-- Register handlers for events, slash commands and timer, etc.
		Apollo.RegisterSlashCommand("kbox", "OnKiruBoxOn", self)
		Apollo.RegisterSlashCommand("kboxtype", "OnKiruBoxType", self)
		Apollo.RegisterSlashCommand("kboxtime", "OnKiruBoxTime", self)
		
		if loadFirst then
			self:CreateWindow(initialType)
			self:SetWindowDefaults(1)
		    self.Windows[1].Window:Show(true, true)

			--start a 1 second timer for refreshes, sync, etc.
			self.Timer = ApolloTimer.Create(1.0, true, "OnTimerEvent", self)
		else
--			Print("KiruBox: Loaded but hidden. Type /kbox when needed.")
		end
	end
end

function KiruBox:CreateWindow(wndType)
	if wndType == nil then wndType = Type_Main end
	local WindowForm = "KiruBox"..wndType
    local wnd = Apollo.LoadForm(self.xmlDoc, WindowForm, nil, self)
	local key = -1
	if wnd ~= nil then
		key = #self.Windows + 1
		self.Windows[key] = {}
		self.Windows[key].Window = wnd
		self.Windows[key].Key = key
		self.Windows[key].Type = wndType
	end
	return key
end

function KiruBox:ReKeyWindow(newIndex)
	--copy the data
	local wnd = self.Windows[newIndex]
	local data = wnd.Window:GetData()
	
	--change it to match the new index
	wnd.Key = newIndex
	wnd.Window:SetData({Key = newIndex})
	wnd.EditBox:SetData({Key = newIndex})
end

function KiruBox:SetWindowDefaults(wndNumber, offset, origin)

	local wnd = self.Windows[wndNumber]
	local wndType = wnd.Type
	
	--set the version number in the title
	wnd.Window:FindChild("Title"):SetText("KiruBox v" .. versionNumber)
	
	wnd.Window:SetData({Key = wndNumber})
	wnd.MaxLength = 498
	wnd.FirstSnip = ""
	wnd.Snips = {}
	wnd.Snips[1] = "I like pie."
	wnd.CurrentSnip = 1
	wnd.SnipTimer = 0
	wnd.NeedsSnip = false
		
	wnd.Current = wnd.Window:FindChild("Current")
	wnd.ClipboardButton = wnd.Window:FindChild("Button_Clipboard")
	wnd.EditBox = wnd.Window:FindChild("EditBox")
	wnd.PreviewArea = wnd.Window:FindChild("PreviewArea")
	wnd.PrepareBox = wnd.PreviewArea:FindChild("Prepare")
	wnd.TotalSnipDisplay = wnd.Window:FindChild("TotalSnips")
	wnd.PickSnipLabel = wnd.Window:FindChild("PickSnip")
	wnd.PickSnipLabel:Show(false,false)
	wnd.ButtonSend = wnd.Window:FindChild("Button_Send")
	wnd.WhisperTargetEdit = wnd.Window:FindChild("EditBox_Whisper") 
	wnd.CustomTargetEdit = wnd.Window:FindChild("EditBox_Custom")
	
	wnd.RadioWhisper = wnd.Window:FindChild("Radio_Whisper")
	wnd.RadioCustom = wnd.Window:FindChild("Radio_Custom")
	wnd.RadioParty = wnd.Window:FindChild("Radio_Party")
	wnd.RadioGuild = wnd.Window:FindChild("Radio_Guild")
	wnd.RadioEmote = wnd.Window:FindChild("Radio_Emote")
	wnd.RadioSay = wnd.Window:FindChild("Radio_Say")

	wnd.SnipButtonBar = {}
	wnd.SnipButtonBar.Window = wnd.Window:FindChild("SnipButtonBar")
	wnd.SnipButtonBar.Buttons = {}

	wnd.PrepareBox:SetText(wnd.Snips[1])
	wnd.EditBox:SetText("")
	wnd.EditBox:SetData({Key = wndNumber})
	wnd.PreviewArea:Show(false,true)
	
	wnd.ContinueEdit = wnd.Window:FindChild("EditBox_ContinueText")
	wnd.PrefixEdit = wnd.Window:FindChild("EditBox_PrefixText")
	wnd.Window:FindChild("Total"):SetText(wnd.MaxLength)
	wnd.ContinueEdit:SetText(self.Continue)
	wnd.PrefixEdit:SetText(self.Prefix)
			
	--set the chat commands to the radio buttons
	wnd.RadioWhisper:SetData({Command = "/w", Label = "Whisper"})
	wnd.RadioCustom:SetData({Command = "CUST", Label = "Custom"})
	wnd.RadioSay:SetData({Command = "/s", Label = "Say"})
	wnd.RadioParty:SetData({Command = "/p", Label = "Party"})
	wnd.RadioGuild:SetData({Command = "/g", Label = "Guild"})
	wnd.RadioEmote:SetData({Command = "/e", Label = "Emote"})
	
	wnd.ButtonRPPrefix = wnd.Window:FindChild("Button_UseRPPrefix")
	
	if wndType == Type_Basic then
		wnd.Options = wnd.Window:FindChild("Options")
		wnd.Options:Show(false,true)
		wnd.Window:SetSizingMinimum(380,164)
	else
		wnd.Window:SetSizingMinimum(512,235)
		wnd.Window:SetSizingMaximum(2512,235)
	end
	
	--throttle stuff
	wnd.Throttle = false
	wnd.ThrottleTime = 5
	wnd.ThrottleSnip = 0
	wnd.ThrottleTimeLeft = 0
	
	--fill in with preferences
	local hasPrefs = false
	if self.Preferences.WindowPos ~= nil and self.Preferences.WindowPos[wndNumber] ~= nil then hasPrefs = true end

	if hasPrefs then
		local prefs = self.Preferences.WindowPos[wndNumber]
		wnd.SendTo = prefs.SendTo
		wnd.WhisperTo = prefs.WhisperTo or "Nobody"
		wnd.CustomTarget = prefs.CustomTarget or "/zone"
		
		wnd.Throttle = prefs.Throttle
		wnd.ThrottleTime = prefs.ThrottleTime or 5
		wnd.UseRPPrefix = prefs.UseRPPrefix or false
	end
	
	wnd.ButtonRPPrefix:SetCheck(wnd.UseRPPrefix == true)
	
	if wnd.Throttle then
		wnd.Window:FindChild("ClockButton"):SetSprite("CRB_BreakoutSprites:spr_BreakoutStun_ClockGreen")
	else
		wnd.Window:FindChild("ClockButton"):SetSprite("CRB_BreakoutSprites:spr_BreakoutStun_ClockOrange")	
	end
		
	wnd.SendTo = wnd.SendTo or "/e"
	wnd.WhisperTo = wnd.WhisperTo or "Nobody"
	wnd.CustomTarget = wnd.CustomTarget or "/zone"
	wnd.WhisperTargetEdit:SetText(wnd.WhisperTo)
	wnd.CustomTargetEdit:SetText(wnd.CustomTarget)
	
	if wnd.SendTo == "/w" then
		wnd.RadioWhisper:GetParent():SetRadioSelButton("SpamChat",wnd.RadioWhisper)
	elseif wnd.SendTo == "/e" then
		wnd.RadioEmote:GetParent():SetRadioSelButton("SpamChat",wnd.RadioEmote)		
	elseif wnd.SendTo == "/p" then
		wnd.RadioParty:GetParent():SetRadioSelButton("SpamChat",wnd.RadioParty)		
	elseif wnd.SendTo == "/s" then
		wnd.RadioSay:GetParent():SetRadioSelButton("SpamChat",wnd.RadioSay)		
	elseif wnd.SendTo == "/g" then
		wnd.RadioGuild:GetParent():SetRadioSelButton("SpamChat",wnd.RadioGuild)		
	else
		wnd.RadioCustom:GetParent():SetRadioSelButton("SpamChat",wnd.RadioCustom)		
	end
	
	self:SetSendButtonText(wnd.ButtonSend, wnd.SendTo)
	
	--check for offset preferences
	if hasPrefs then
			--restore offsets only if the types match
		if self.Preferences.WindowPos[wndNumber].Type == wnd.Type then		
			local RestoreAnchors = self.Preferences.WindowPos[wndNumber]
			if RestoreAnchors then
				wnd.Window:SetAnchorOffsets(RestoreAnchors.l, RestoreAnchors.t, RestoreAnchors.r, RestoreAnchors.b)
			end
		end
	else
		if offset ~= nil then
			local l,t,r,b = wnd.Window:GetAnchorOffsets()
			
			if origin ~= nil then
				--shift B and R over by T + height, L + width
				b = origin.t + (b - t)
				r = origin.l + (r - l)
				--shift L and T over
				l = origin.l
				t = origin.t
			end
			
			if offset.h ~= nil then 
				l = l + offset.h 
				r = r + offset.h
			end
			if offset.v ~= nil then 
				t = t + offset.v 
				b = b + offset.v
			end
			wnd.Window:SetAnchorOffsets(l,t,r,b)
		end
	end
end

function KiruBox:SetSendButtonText(button,command,special)
	if special then button:SetText(special) return end
	
	if command == "/w" then
		button:SetText("Send All to Whisper")
	elseif command == "/e" then
		button:SetText("Send All to Emote")
	elseif command == "/p" then
		button:SetText("Send All to Party")
	elseif command == "/s" then
		button:SetText("Send All to Say")
	elseif command == "/g" then
		button:SetText("Send All to Guild")
	else
		button:SetText("Send All to Custom")
	end
end

function KiruBox:FindParentWindowID(parent)
	local key = -1
	--if parent is nil, return its key
	if parent == nil then return key end		
	--loop through parents until we get -1 back
	key = self:FindParentWindowID(parent:GetParent())
	
	if key == -1 then
		--no parent key, use ours
		key = parent:GetData().Key
	end
	
	return key
end

-----------------------------------------------------------------------------------------------
-- Settings Persistance
-----------------------------------------------------------------------------------------------

function KiruBox:OnSave(eType)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
	
	local saveData = {}	

	--save open window positions
	saveData.Windows = {}
	--if there's no open windows, save the first one
	if #self.Windows == 0 then
		saveData.LoadFirst = false
		local backup = self.Preferences.WindowPos[1]
		saveData.Windows[1] = { 
			l = backup.l, 
			t = backup.t, 
			r = backup.r, 
			b = backup.b, 
			Type = backup.Type, 
			SendTo = backup.SendTo, 
			WhisperTo = backup.WhisperTo, 
			CustomTarget = backup.CustomTarget,
			Throttle = backup.Throttle,
			ThrottleTime = backup.ThrottleTime,
			UseRPPrefix = backup.UseRPPrefix }			
	else
		saveData.LoadFirst = true
		--else save all of them
		for key, win in ipairs(self.Windows) do
			local l,t,r,b = win.Window:GetAnchorOffsets()
			if win.Type == Type_Main and win.PreviewArea:IsVisible() then b = b - 140 end
			saveData.Windows[key] = { 
				l = l, 
				t = t, 
				r = r, 
				b = b, 
				Type = win.Type, 
				SendTo = win.SendTo, 
				WhisperTo = win.WhisperTo, 
				CustomTarget = win.CustomTarget,
				Throttle = win.Throttle,
				ThrottleTime = win.ThrottleTime,
				UseRPPrefix = win.UseRPPrefix }			
		end
	end	
	--save other stuff
	saveData.ContinueText = self.Continue
	saveData.PrefixText = self.Prefix
	saveData.WindowType = self.Preferences.WindowType
	
	return saveData
end

function KiruBox:OnRestore(eType, tSavedData)
	if eType ~= GameLib.CodeEnumAddonSaveLevel.Character then
		return
	end
			
	self.FromRestore = tSavedData	
end

function KiruBox:OnTimerEvent()
	--check each window for snip events
	for key, win in ipairs(self.Windows) do
		if win.NeedsSnip then
			if win.SnipTimer == 0 then
				self:ProcessSnips(key)
			else
				win.SnipTimer = win.SnipTimer - 1
			end			
		end
		self:UpdatePrepareBox(win.CurrentSnip, key)
		
		--any throttling to do?
		if win.ThrottleTimeLeft > 0 then
			win.ThrottleTimeLeft = win.ThrottleTimeLeft - 1
			if win.ThrottleTimeLeft == 0 then
				--send the next snip
				self:SendSnip(win.ChatCommand, win.Snips[win.ThrottleSnip])
				local numSnips = #win.Snips
				if numSnips > win.ThrottleSnip then
					--reset for the next snip
					win.ThrottleSnip = win.ThrottleSnip + 1
					win.ThrottleTimeLeft = win.ThrottleTime
				else
					--done throttling, turn stuff back on
					self:EnableUIThrottle(win, true)
					win.ButtonSend:Enable(true)
				end
			else
			--change the button text
				self:SetSendButtonText(win.ButtonSend, win.SendTo, "Sending Snip " .. win.ThrottleSnip .. " in " .. win.ThrottleTimeLeft)
			end
		end
	end --end for
end

function KiruBox:EnableUIThrottle(window, enable)
	--enable or disable stuff while throttle-sending
	local enableColor = "FFFFFFFF"
	if enable then
		self:SetSendButtonText(window.ButtonSend, window.SendTo)
	else
		enableColor = "FF777777"
		window.ButtonSend:SetText("Sending Snip 2 in " .. window.ThrottleTime)				
	end
	window.RadioWhisper:SetTextColor(enableColor)
	window.RadioCustom:SetTextColor(enableColor)
	window.RadioParty:SetTextColor(enableColor)
	window.RadioGuild:SetTextColor(enableColor)
	window.RadioEmote:SetTextColor(enableColor)
	window.RadioSay:SetTextColor(enableColor)

	window.ButtonSend:Enable(enable)
	window.EditBox:Enable(enable)
	window.RadioWhisper:Enable(enable)
	window.RadioCustom:Enable(enable)
	window.RadioParty:Enable(enable)
	window.RadioGuild:Enable(enable)
	window.RadioEmote:Enable(enable)
	window.RadioSay:Enable(enable)
	window.ContinueEdit:Enable(enable)
	window.PrefixEdit:Enable(enable)
	window.WhisperTargetEdit:Enable(enable)
	window.CustomTargetEdit:Enable(enable)
end

-- on SlashCommand "/kbox"
function KiruBox:OnKiruBoxOn()
	if self.Timer == nil then self.Timer = ApolloTimer.Create(1.0, true, "OnTimerEvent", self) end
	--make a new window
	--first check if the 'next' window is in our preferences
	local prefs = self.Preferences.WindowPos
	local nextKey = #self.Windows + 1
	local newType
	if prefs ~= nil and prefs[nextKey] ~= nil then newType = prefs[nextKey].Type end
	if newType == nil then newType = self.Preferences.WindowType end
	
	local key = self:CreateWindow(newType)
	if key > 0 then self:SetWindowDefaults(key) end
end

-- on SlashCommand "/kboxtype"
function KiruBox:OnKiruBoxType()
	if self.Preferences.WindowType == Type_Basic then
		self.Preferences.WindowType = Type_Main
		Print("KiruBox: New Boxes will be of Type: Normal")
	else
		self.Preferences.WindowType = Type_Basic
		Print("KiruBox: New Boxes will be of Type: Simple")
	end
end

function KiruBox:ClearText(winKey)
	if self.Windows[winKey] == nil then return end
	local wnd = self.Windows[winKey]
	--if we're in the middle of a throttle, don't do anything.
	if wnd.Throttle and wnd.ThrottleTimeLeft and wnd.ThrottleTimeLeft > 0 then return nil end
	wnd.EditBox:SetText("")
	wnd.NeedsSnip = false
	wnd.Current:SetText("0")
	wnd.Current:SetTextColor("FF00FF00")
	self:ClearSnipButtonBar(wnd.SnipButtonBar)
	wnd.CurrentSnip = 1
	wnd.Snips = {}
	wnd.Snips[1] = ""
	wnd.TotalSnipDisplay:SetText(1)
	wnd.PickSnipLabel:Show(false,true)
	wnd.EditBox:SetFocus()
end

function KiruBox:EditBox_Changed( wndHandler, wndControl, strText )
	--get which window we're in
	local wndKey = wndControl:GetData().Key
	if wndKey == nil then wndKey = self:FindParentWindowID(wndControl:GetParent()) end
	if wndKey == nil then wndControl:SetText("Error in code. Call for Kiru.") return end

--	if strText:find("\n") ~= nil then
--		strText = self:StripReturn(strText)
--		Print(strText)
--		Print(tostring(wndControl:IsStyleExOn("NoSelectOnFocus")))
--		wndControl:SetStyleEx("NoSelectOnFocus")
--		Print(tostring(wndControl:IsStyleExOn("NoSelectOnFocus")))
--		wndControl:SetText(strText)
--		Print(tostring(wndControl:IsStyleExOn("NoSelectOnFocus")))
--		wndControl:SetStyleEx("NoSelectOnFocus",false)
--		Print(tostring(wndControl:IsStyleExOn("NoSelectOnFocus")))
--	end
	
	local wnd = self.Windows[wndKey]
	local useRPPrefix = wnd.UseRPPrefix
	local sLen = #strText
	if useRPPrefix == true then sLen = sLen + #self.RPPrefix end
	
	wnd.Current:SetText(sLen)
	if sLen > wnd.MaxLength then
		wnd.NeedsSnip = true
		wnd.SnipTimer = 1
		wnd.Current:SetTextColor("FFFF0000")
	else
		wnd.NeedsSnip = false
		wnd.Current:SetTextColor("FF00FF00")
		self:ClearSnipButtonBar(wnd.SnipButtonBar)
		wnd.CurrentSnip = 1
		wnd.Snips = {}
		if useRPPrefix == true then strText = self.RPPrefix .. strText end
		wnd.Snips[1] = strText
		wnd.TotalSnipDisplay:SetText(1)
		wnd.PickSnipLabel:Show(false,true)
	end	
end

function KiruBox:ClearSnipButtonBar(barWnd)
	if #barWnd.Buttons > 0 then 
		barWnd.Window:DestroyChildren()
		barWnd.Buttons = {}
		barWnd.Window:ArrangeChildrenVert()
	end	
end

function KiruBox:ProcessSnips(wndKey)
	local wnd = self.Windows[wndKey]
	local useRPPrefix = wnd.UseRPPrefix
	
	local strText = self:Trim(wnd.EditBox:GetText())
	if useRPPrefix == true then strText = self.RPPrefix .. strText end
	local sLen = #strText
	
	wnd.Snips = {}
	--if the string is under the max, we're done.
	if sLen < wnd.MaxLength then
		wnd.Snips[1] = strText
		wnd.FirstSnip = wnd.Snips[1]
		wnd.NeedsSnip = false
		self:ClearSnipButtonBar(wnd.SnipButtonBar)	
		wnd.TotalSnipDisplay:SetText(1)
		wnd.PickSnipLabel:Show(false,true)
		return
	end
	
	local snipCount = 1
	--break up the text into individual snips, taking into account we might be adding a Continue string
	local maxLength = wnd.MaxLength - self.ContinueLen
	
	--loop through the string until there is no more huge strings
	while sLen > maxLength do
		--look at the character right after the max length  
		local lastChar = string.sub(strText, maxLength + 1)
		
		--if it's a space, then we're good on this snip.
		if lastChar == " " then
			wnd.Snips[snipCount] = string.sub(strText, 1, maxLength)
			strText = self:Trim(string.sub(strText, maxLength + 1))
			--if there is anything left, add the prefix to it so it becomes part of the text
			if useRPPrefix == true then strText = self.RPPrefix .. strText end
			if self.PrefixLen > 0 then strText = self.Prefix .. strText end
			sLen = #strText
		else
			--we need to find a breaking point.
			--take the maximum
			local snip = string.sub(strText, 1, maxLength)
			--count backwards till we find a space
			local breakPoint = string.find(snip,"%s[^%s]*$")	
			--if no space found, try a hyphen
			if breakPoint == nil then
				breakPoint = string.find(snip,"[-]*$")
			end
			--if still nothing, we're breaking at the end
			if breakPoint == nil then breakPoint = maxLength end
			local addsnip = string.sub(strText, 1, breakPoint)
			if self.ContinueLen > 0 then addsnip = addsnip .. self.Continue end
			wnd.Snips[snipCount] = addsnip
			strText = self:Trim(string.sub(strText, breakPoint + 1))
			--if there is anything left, add the prefix to it so it becomes part of the text
			if useRPPrefix == true then strText = self.RPPrefix .. strText end
			if self.PrefixLen > 0 then strText = self.Prefix .. strText end
			sLen = #strText		
		end
		snipCount = snipCount + 1			
	end
	--last snip is whatever is left
	wnd.Snips[snipCount] = strText
	wnd.FirstSnip = wnd.Snips[1]
	wnd.NeedsSnip = false
	
	self:MakeSnipButtons(wnd)
	wnd.SnipButtonBar.Window:ArrangeChildrenHorz()
	wnd.PickSnipLabel:Show(true,true)
end

function KiruBox:MakeSnipButtons(wnd)
	self:ClearSnipButtonBar(wnd.SnipButtonBar)
	
	local numberOfSnips = #wnd.Snips
	if numberOfSnips < 2 then return end
	for i = 1, numberOfSnips, 1 do
		local wndButton = Apollo.LoadForm(self.xmlDoc, "SnipButton", wnd.SnipButtonBar.Window, self)
		wndButton:SetData({Snip = i})
		wndButton:SetText(i)
		wndButton:Enable(i ~= wnd.CurrentSnip)
		wnd.SnipButtonBar.Buttons[i] = wndButton
	end
	wnd.TotalSnipDisplay:SetText(numberOfSnips)
end

function KiruBox:UpdatePrepareBox(snipNumber, wndKey)
	--just in case the Snip got cleared out between the time the timer fired and now
	local updateText = self.Windows[wndKey].Snips[snipNumber]
	if updateText ~= nil then self.Windows[wndKey].PrepareBox:SetText(updateText) end
end

function KiruBox:PrepareSnip( wndHandler, wndControl, x, y )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	
	local wnd = self.Windows[key]
	local copyText = wnd.PrepareBox:GetText()
	
	if copyText ~= nil and copyText ~= "" then
		wnd.ClipboardButton:SetActionData(GameLib.CodeEnumConfirmButtonType.CopyToClipboard, copyText)
	end
end

function KiruBox:Snip_Button_Pushed( wndHandler, wndControl, eMouseButton )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end

	local wnd = self.Windows[key]
	local snipNumber = wndControl:GetData().Snip
	
	wnd.CurrentSnip = snipNumber
	self:UpdatePrepareBox(snipNumber, key)

	for i = 1, #wnd.SnipButtonBar.Buttons, 1 do
		wnd.SnipButtonBar.Buttons[i]:Enable(i ~= snipNumber)
	end
end

function KiruBox:Button_SendAll( wndHandler, wndControl, eMouseButton )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return nil end
	if self.Windows[key] == nil then return nil end
	local wnd = self.Windows[key]

	local chatCommand = self:CreateChatCommand(key)
	
	--do we send all at once, or throttle it?
	if wnd.Throttle then
		--send the first snip now
		self:SendSnip(chatCommand, wnd.Snips[1])
		
		--if there's multiple snips set up the timer
		if #wnd.Snips > 1 then
			wnd.ChatCommand = chatCommand
			wnd.ThrottleTimeLeft = wnd.ThrottleTime
			wnd.ThrottleSnip = 2
			self:EnableUIThrottle(wnd, false)
		end	
	else
		for key,val in ipairs(wnd.Snips) do
			self:SendSnip(chatCommand, val)
		end	
	end
end

function KiruBox:SendSnip(chatCommand, snipText)
	ChatSystemLib.Command(chatCommand .. " " .. snipText)
end

function KiruBox:CreateChatCommand(windowKey)

	local wnd = self.Windows[windowKey]

	local chatCommand = wnd.SendTo
	if chatCommand == nil or chatCommand == "" then 
		Print("KiruBox: Please choose where to send the messages.") 
		return nil 
	end

	if wnd.SendTo == "/w" then 
		if wnd.WhisperTo == nil or wnd.WhisperTo == "" then 
			Print("KiruBox: Please enter a Whisper recipient.") 
			return nil 
		end
		chatCommand = chatCommand .. " " .. wnd.WhisperTo
	end
	
	if wnd.SendTo == "CUST" then
		if wnd.CustomTarget == nil or wnd.CustomTarget == "" then
			Print("KiruBox: You chose a Custom channel but did not enter one.  Please do so.")
			return nil
		end
		chatCommand = wnd.CustomTarget
	end
	
	return chatCommand
end

function KiruBox:Radio_Spam( wndHandler, wndControl, eMouseButton )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	local data = wndControl:GetData()
	local commandLabel = data.Label
	self.Windows[key].ButtonSend:SetText("Send All to " .. commandLabel)
	self.Windows[key].SendTo = data.Command
	--if whisper, make its edit box in focus
	if data.Command == "/w" then
		self.Windows[key].WhisperTargetEdit:SetFocus()
	end	
	--if its custom, make its edit box in focus
	if data.Command == "CUST" then
		self.Windows[key].CustomTargetEdit:SetFocus()
	end
end

function KiruBox:EditBox_Whisper_Changed( wndHandler, wndControl, strText )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	self.Windows[key].WhisperTo = strText
	if not self.Windows[key].RadioWhisper:IsChecked() then 
		wndControl:GetParent():SetRadioSelButton("SpamChat",self.Windows[key].RadioWhisper)
		self.Windows[key].ButtonSend:SetText("Send All to Whisper")
	end
end

function KiruBox:EditBox_Continue_Changed( wndHandler, wndControl, strText )
	self.Continue = strText
	self.ContinueLen = #strText
	
	--trigger a snip update if the windows have more than one snip in them
	for key, win in pairs(self.Windows) do
		if #win.Snips > 1 then
			win.SnipTimer = 1
			win.NeedsSnip = true
		end
		if win.ContinueEdit ~= wndControl then
			--also set their continue box to match
			win.ContinueEdit:SetText(strText)
		end
	end
end

function KiruBox:EditBox_Prefix_Changed( wndHandler, wndControl, strText )
	self.Prefix = strText
	self.PrefixLen = #strText
	
	--trigger a snip update if the windows have more than one snip in them
	for key, win in pairs(self.Windows) do
		if #win.Snips > 1 then
			win.SnipTimer = 1
			win.NeedsSnip = true
		end
		
		if win.PrefixEdit ~= wndControl then
			--also set their prefix box to match
			win.PrefixEdit:SetText(strText)
		end
	end
end

function KiruBox:EditBox_Custom_Changed( wndHandler, wndControl, strText )
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	self.Windows[key].CustomTarget = strText
	if not self.Windows[key].RadioCustom:IsChecked() then 
		wndControl:GetParent():SetRadioSelButton("SpamChat",self.Windows[key].RadioCustom)
		self.Windows[key].ButtonSend:SetText("Send All to Custom")
	end
end

function KiruBox:Button_TogglePreview( wndHandler, wndControl, eMouseButton )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	local wnd = self.Windows[key]

	local toggle = wnd.PreviewArea:IsVisible()
	local l,t,r,b,newBottom = 0
	if wnd.Type == Type_Main then
		--need to do some window resizing
		l,t,r,b = wnd.Window:GetAnchorOffsets()
		--if it was hiding, we need to stretch out the main window
		newBottom = b + 140
		--if it was already showing, gotta shrink instead
		if toggle then newBottom = b - 140 end
		
		local resize = 235
		if not toggle then resize = 235+140 end
		
		wnd.Window:SetSizingMinimum(512,resize)
		wnd.Window:SetSizingMaximum(2512,resize)
	end	

	if toggle then 
		wndControl:SetText("Show Preview")
	else
		wndControl:SetText("Hide Preview")	
	end	
	wnd.PreviewArea:Show(not toggle,true)
	if wnd.Type == Type_Main then wnd.Window:SetAnchorOffsets(l,t,r,newBottom) end
end

function KiruBox:Button_ToggleOptions( wndHandler, wndControl, eMouseButton )
	--get parent window key
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	local wnd = self.Windows[key]

	if wnd.Options == nil then return end
	
	local toggle = wnd.Options:IsVisible()
	wnd.Options:Show(not toggle, true)
	if not toggle then wnd.Window:BringChildToTop(wnd.Options) end
end

function KiruBox:Trim(input)
	--remove whitespace from either end of string
	return (input:gsub("^%s*(.-)%s*$", "%1"))
end

function KiruBox:LeftTrim(input)
	-- remove leading whitespace from string.
  	return (input:gsub("^%s*", ""))
end

function KiruBox:StripReturn(input)
	-- remove leading whitespace from string.
  	return (input:gsub("\n+", " "))
end

function KiruBox:RightTrim(input)
	-- remove trailing whitespace from string.
	local n = #input
	while n > 0 and input:find("^%s", n) do n = n - 1 end
	return input:sub(1, n)
end

---------------------------------------------------------------------------------------------------
-- KiruBoxMain Functions
---------------------------------------------------------------------------------------------------

function KiruBox:Button_Close( wndHandler, wndControl, eMouseButton )
	local countWin = #self.Windows
	if countWin > 1 then
		--need to renumber all the other keys if this one happens to not be the last one created
		--get parent window key
		local key = self:FindParentWindowID(wndControl:GetParent())
		if key == nil or key == -1 then return end
		self.Windows[key].Window:Destroy()
		table.remove(self.Windows,key)
		if key ~= countWin then
			--this wasn't is the last window, need to renumber the other keys after we nuke this one
			for newIndex = key, countWin - 1, 1 do
				self:ReKeyWindow(newIndex)
			end
		end
	else
		--this is the last window.. advise user how to reopen.	
		--Print("KiruBox: Closed. Type /kbox to bring it back.")
		--turn off the timer
		self.Timer = nil
		
		local win = self.Windows[1]
		--store it in Preferences.WindowPos for later retrieving 
		if self.Preferences.WindowPos == nil then self.Preferences.WindowPos = {} end
		local l,t,r,b = win.Window:GetAnchorOffsets()
		if win.Type == Type_Main and win.PreviewArea:IsVisible() then b = b - 140 end
		self.Preferences.WindowPos[1] = { l = l, t = t, r = r, b = b, Type = win.Type, SendTo = win.SendTo, WhisperTo = win.WhisperTo, CustomTarget = win.CustomTarget }			
		
		--finally, get rid of it
		win.Window:Destroy()
		self.Windows = {}
	end
end

function KiruBox:Button_Add( wndHandler, wndControl, eMouseButton )
	--make a new window near this one
	local prefs = self.Preferences.WindowPos
	local nextKey = #self.Windows + 1
	local newType
	if prefs ~= nil and prefs[nextKey] ~= nil then newType = prefs[nextKey].Type end
	if newType == nil then newType = self.Preferences.WindowType end	
	
	local newkey = self:CreateWindow(newType)
	local offset = {h = 50, v = 50}
	local l,t,r,b = wndControl:GetParent():GetAnchorOffsets()
	local origin = {l = l, t = t}
	
	if newkey > 0 then self:SetWindowDefaults(newkey, offset, origin) end
end

---------------------------------------------------------------------------------------------------
-- KiruBoxStandard Functions
---------------------------------------------------------------------------------------------------

function KiruBox:Button_Clear( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY )
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	self:ClearText(key)
end

function KiruBox:ToggleThrottle( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	local window = self.Windows[key]	
	local throttle = window.Throttle
	window.Throttle = not throttle
	if window.Throttle then
		wndControl:SetSprite("CRB_BreakoutSprites:spr_BreakoutStun_ClockGreen")
	else
		wndControl:SetSprite("CRB_BreakoutSprites:spr_BreakoutStun_ClockOrange")	
	end
end

function KiruBox:Button_UseRPPrefix( wndHandler, wndControl, eMouseButton )
	local key = self:FindParentWindowID(wndControl:GetParent())
	if key == nil or key == -1 then return end
	self.Windows[key].UseRPPrefix = wndControl:IsChecked()
end

-----------------------------------------------------------------------------------------------
-- KiruBox Instance
-----------------------------------------------------------------------------------------------
local KiruBoxInst = KiruBox:new()
KiruBoxInst:Init()
